"""
ALP2-Team
SS 2012

Solutions for the Tasks 1-3 of the 3th homework
(1. 2. und 3. Aufgabe des 4. Übungsblattes)
"""

# Task 1 (1. Aufgabe)

from time import clock
from math import isnan
from random import randint, random


""" Help functions """

def factorial_end_rec(n):
    """ Calculate n! with an end recursive help function which is only defined for natural numbers.
        We use an internal function definition because it only makes sense to use it in the context
        of the outer factorial function. The function need to be always call with the constant 1 as
        its first argument.
    """
    def factorial_helper (fact, n):
        """ fact acumulate the multiplications and n is use to make
            the count down for stopping the recursion, when it becomes 0 """
        if n==0:
            return fact
        else:
            return factorial_helper (fact*n, n-1)
        
    return factorial_helper (1, n)

def factorial_iter(n):
    """ Calculate n! using an iterative algorithm and returns 0
        for negative n values.
    """
    if n<0: 
        return 0           
    fact = 1
    for i in range(1, n+1):      
        fact *= i   
    return fact


def binom_naiv(n,k):
    """ This function implements a naive algorithm to calculate
        the binomial coefficient using a non recursive version of the
        factorial function
    """
    if 0<=k<=n:
        return factorial_iter(n)//(factorial_iter(k)*factorial_iter(n-k))
    else:
        return 0
    

def binom_naiv_rec(n, k):
    """ This function calculate the binomial coefficient with
        the standard recursive definition of the function
    """
    if 0<=k<=n:
        if k==0 or k==n:
            return 1
        else:
            if k==1 or k==(n-1):
                return n
            else:
                return binom_naiv_rec(n-1, k) + binom_naiv_rec(n-1, k-1)
    else:
        return 0

def binom_iter(n,k):
    """ This is a much more effizient iterative Algorithm to calculate the binomial
        coefficient. It use exactly the definition of the homework """
    if 0<=k<=n:
        if k==0:
            return 1
        else:
            nominator, denominator = 1, 1
            for i in range(1,k+1):
                nominator *= (n-i+1)
                denominator *= i
            return nominator//denominator
    else:
        return 0

def test_binom_funcs():
    """ test some values of the binomial coefficient """
    print("binom_iter(0,0)=",binom_iter(0,0))
    print("binom_naiv_rec(0,0)=",binom_naiv_rec(0,0))
    print("binom_naiv_rec(0,0)=",binom_naiv(0,0))

    print("binom_iter(1,3)=",binom_iter(1,3))
    print("binom_naiv_rec(1,3)=",binom_naiv_rec(1,3))
    print("binom_naiv(1,3)=",binom_naiv(1,3))

    print("binom_iter(10,0)=",binom_iter(10,0))
    print("binom_naiv_rec(10,0)=",binom_naiv_rec(10,0))
    print("binom_naiv(10,0)=",binom_naiv(10,0))

    print("binom_iter(100,3)=",binom_iter(100,3))
    print("binom_naiv_rec(100,3)=",binom_naiv_rec(100,3))
    print("binom_naiv(100,3)=",binom_naiv(100,3))

def compare_binom_funcs():
    """ Here we make a pragmatic comparition of the three different implementation
        of the binomial coefficient functions. We calculate all posible values of
        k and n between 1 and 20 and repeat the calculations 20 times for each function.
    """
    z0 = clock()
    for i in range(1,20):
        for j in range(1,20):
            for k in range(j):
                binom_naiv(j,k)
    z1 = clock()
    print("time of binom_naiv = " + str(z1-z0))

    z2 = clock()
    for i in range(1,20):
        for j in range(1,20):
            for k in range(j):
                binom_naiv_rec(j,k)
    z3 = clock()
    print("time of binom_naiv_rec = " + str(z3-z2))

    z4 = clock()
    for i in range(1,20):
        for j in range(1,20):
            for k in range(j):
                binom_iter(j,k)
    z5 = clock()
    print("time of binom_naiv_iter = " + str(z5-z4))
    

# Task 2 (2. Aufgabe)
""" The power functions of this solutions are defined for every integer exponent
    and any number as basis with the only exception that 0**exp for exp<= 0 is not defined
    and returns isnan (is not a number). This is different from the python power-Operator
    which returns 1 for 0**0 and return an exception for negativ exponents of zero
    """
def power_rec(basis,exp):
       
    def help_power(result, basis, exp):
        """ An end recursive help function is used for the calculations """ 
        if exp>0:
            return help_power(result*basis, basis, exp-1)
        else:
            return result

    if basis == 0:
        if exp <= 0:
            return isnan
        else:
            return 0
    if exp<0:
        return 1/help_power(1, basis, -exp)
    else:
        return help_power(1, basis, exp)
    

def power_iter(basis, exp):
    if basis == 0:
        if exp <= 0:
            return isnan
        else:
            return 0
    result = 1
    if exp<0:
        for i in range(-exp):
            result *= basis
        return 1/result
    else:
        for i in range(exp):
            result *= basis
        return result

def test_power_functions():
    """ Test some spezial cases """
    t00 = power_iter(0,0) == power_rec(0,0) and  power_iter(0,0) == isnan
    t01 = power_iter(0,1) == power_rec(0,1) and  power_iter(0,1) == 0**1
    t10 = power_iter(1,0) == power_rec(1,0) and  power_iter(1,0) == 1**0
    t0_1 = power_iter(0,-1) == power_rec(0,-1) and  power_iter(0,-1) == isnan
    t3_0 = power_iter(3,0) == power_rec(3,0) and  power_iter(3,0) == 3**0
    cmp = t00 and t01 and t10 and t0_1 and t3_0
    if cmp:
        print("the calculation of the spezial cases is correct")
    else:
        print("some of the functions has errors")
    """ Test with random integer numbers """
    cmp = True
    for i in range(1,10):
        b = randint(1,10)
        e = randint(1,10)
        cmp = cmp and power_iter(b,e)==power_rec(b,e) and  power_iter(b,e)==(b**e)
    if cmp:
        print("The calculations of the power function with random basis are apparently correct")
    cmp = True
    """ Test with a random flotting point basis and a random integer exponent.
        Because of the rounding errors, we can't compare floating point numbers
    """
    print("The followings are calculations of both power functions with floating point basis")
    for i in range(1,10):
        b = random()*10+1
        e = randint(1,10)
        print(power_iter(b,e),power_iter(b,e))
    print(cmp)
        

# Task 3 (3. Aufgabe)

def bauer_mult(a,b):
    """ this rusian multiplication is only defined for positive numbers """
    mult = 0
    while a>0:
        if a&1:         # test of the last bit is equal 1 (odd number)
            mult += b
        b <<= 1         # equivalent to b**2
        a >>= 1         # equivalent to b//2
    return mult

def bauer_mult_int(a,b):
    """ this rusian multiplication is defined for integer numbers """
    if a<0:
        return -bauer_mult(-a, b)
    else:
        return bauer_mult(a, b)

def test_bauer_mult():
	print("bauer_mult(10,70)=",bauer_mult(10,70))
	print("bauer_mult(100,75)=",bauer_mult(100,75))
	print("bauer_mult(10,-7)=",bauer_mult(10,-7)) # only the second operand can be negative in bauer_mult
	print("bauer_mult_int(-20,7)=",bauer_mult_int(-20,7))   
	print("bauer_mult_int(-20,0)=",bauer_mult_int(-20,0))
	print("bauer_mult_int(-20,-20)=",bauer_mult_int(-20,-20)) # both operands can be negative in bauer_mult_int
	print("bauer_mult_int(0,-20)=",bauer_mult_int(0,-20))
	print("bauer_mult_int(0,0)=",bauer_mult_int(0,0))
	print("bauer_mult_int(1,-20)=",bauer_mult_int(1,-20))
	print("bauer_mult_int(121,-11)=",bauer_mult_int(121,-11))

def test_all_tasks():
    print("\n--- Test of some values of the binomial coefficient ")
    test_binom_funcs()
    print("\n--- Comparison of the performance of the binomial coefficient functions")
    compare_binom_funcs()
    print("\n--- Test of the power functions for some spezial cases und for random values")
    test_power_functions()
    print("\n--- Test of the russian multiplication")
    test_bauer_mult()

test_all_tasks()
    



